-- debug_status = 1
debug_mod_name = "BlackMarket"
debug_file = debug_mod_name .. "-debug.txt"
prices_file = debug_mod_name .. "-prices.csv"
specials_file = debug_mod_name .. "-specials.txt"
require("utils")
require("config")

--------------------------------------------------------------------------------------
function format_money( n )
	local neg, mega
	
	if n > 1e12 then
		n = math.floor(n/1e6)
		mega = true
	else
		mega = false
	end
	
	if n < 0 then
		n = -n
		neg = true
	else
		neg = false
	end
	
	local s = tostring(math.floor(n+0.5))
	local s2 = ""
	local l = string.len(s)
	local i = l+1
	
	while i > 4 do
		i = i-3	
		s2 =  thousands_separator .. string.sub(s,i,i+2) .. s2
	end
	
	if i > 1 then
		s2 =  string.sub(s,1,i-1) .. s2
	end
	
	if mega then
		s2 = s2 .. "Mu"
	else
		s2 = s2 .. "u"
	end
	
	if neg then
		return( "-" .. s2 )
	else
		return( s2 )
	end
end

--------------------------------------------------------------------------------------
local function init_tax_rates()
	global.tax_rates = {}
	
	for _, period in ipairs(periods) do
		if period == 0 then
			global.tax_rates[period] = tax_immediate -- tax for immediate action
		else
			global.tax_rates[period] = math.floor(0.5+tax_start * ((24/period) ^ tax_growth))
		end
	end
	
	for period,tax in pairs(global.tax_rates) do
		debug_print( "tax ", period, "h = ", tax, "%" )
	end
end

--------------------------------------------------------------------------------------
local function update_tech_cost(tech)
	local cost = 0
	
	cost = cost + #tech.research_unit_ingredients * tech.research_unit_count * tech_ingr_cost
	cost = cost + tech.research_unit_energy * energy_cost
	
	global.techs_costs[tech.name] = cost
	
	for _, effect in pairs(tech.effects) do
		if effect.type == "unlock-recipe" then
			global.recipes_tech[effect.recipe] = tech.name
		end
	end
end

--------------------------------------------------------------------------------------
local function update_techs_costs()
	global.techs_costs = {}
	global.recipes_tech = {}
	
	for name, tech in pairs(game.forces.player.technologies) do
		update_tech_cost(tech)
	end
end

--------------------------------------------------------------------------------------
local function list_techs_costs()
	debug_print("--------------------------------------------------------------------------------------")
	debug_print("list_techs_costs")
	
	for name, cost in pairs(global.techs_costs) do
		debug_print("tech ", name, " = ", cost)
	end
end

--------------------------------------------------------------------------------------
local function update_objects_prices_start()
	if global.prices_loop > -1 then return end
	
	-- debug_print("--------------------------------------------------------------------------------------")
	-- debug_print("update_objects_prices_start")
	
	debug_print("RAZ")
	
	global.prices = {}
	global.orig_resources = {}
	global.new_resources = {}
	global.free_products = {}
	global.unknowns = {}
	
	global.recipes_used = {}
	global.recipes_parsed = {}
	
	local ingredients_to_parse = {}
	local orig_resources = {}
		
	-- special objects
	
	for name, price in pairs(special_prices) do
		global.prices[name] = {overall=price, tech=0, ingrs=0, energy=0}
		ingredients_to_parse[name] = true
	end

	-- official resources
	
	for name, ent in pairs(game.entity_prototypes) do
		if ent.type == "resource" then 
			if ent.resource_category == "basic-solid" or ent.resource_category == "basic-fluid" then
				-- debug_print( "resource = ", name )
				global.prices[name] = {overall=resource_price, tech=0, ingrs=0, energy=0}
				ingredients_to_parse[name] = true
				orig_resources[name] = true
			end
		end
	end
	
	-- items made out of no ingredients or no recipes
	-- mark item "free" (recipe with no ingredients) or "regularly" produced or consumed
	-- the rest will be items with no recipe at all
	
	local free_products = {}
	local regular_products = {}
	
	for name_recipe, recipe in pairs(game.forces.player.recipes) do
		if string.sub(name_recipe,1,5) ~= "repl-" then
			if recipe.ingredients == nil or #recipe.ingredients == 0 then
				for _, prod in pairs(recipe.products) do
					if global.prices[prod.name] == nil then
						free_products[prod.name] = true -- mark free product
						debug_print( "free ", prod.name, " recipe=", name_recipe)
					end
				end
			else
				for _, prod in pairs(recipe.products) do
					regular_products[prod.name] = true -- mark regular not-free product
				end
			end
		end
	end
	
	-- mark potential undeclared new resources (that are ingredients but never produced)
	
	local new_resources = {}
	
	for _, recipe in pairs(game.forces.player.recipes) do
		if recipe.ingredients ~= nil then
			for _, ingr in pairs(recipe.ingredients) do
				if global.prices[ingr.name] == nil and regular_products[ingr.name] == nil and free_products[ingr.name] == nil then
					new_resources[ingr.name] = true -- mark as possible resource
					ingredients_to_parse[ingr.name] = true
				end
			end
		end
	end
	
	-- give price to free/new resources/unknown

	debug_print("--------------------------------------------------------------------------------------")
	debug_print( "set free/newres/unknown")

	for name, object in pairs(game.item_prototypes) do
		if global.prices[name] == nil then
			if new_resources[name] == true then
				debug_print( "set ", name, " resource_price_new")
				global.prices[name] = {overall=resource_price_new, tech=0, ingrs=0, energy=0}
				ingredients_to_parse[name] = true
			elseif free_products[name] == true then
				debug_print( "set ", name, " free_price")
				global.prices[name] = {overall=free_price, tech=0, ingrs=0, energy=0}
			elseif regular_products[name] == nil then
				debug_print( "set ", name, " unknown_price")
				global.prices[name] = {overall=unknown_price, tech=0, ingrs=0, energy=0}
				global.unknowns[name] = true
			end
		end
	end
	
	for name, object in pairs(game.fluid_prototypes) do
		if global.prices[name] == nil then
			if new_resources[name] == true then
				debug_print( "set ", name, " resource_price_new")
				global.prices[name] = {overall=resource_price_new, tech=0, ingrs=0, energy=0}
				ingredients_to_parse[name] = true
			elseif free_products[name] == true then
				debug_print( "set ", name, " free_price")
				global.prices[name] = {overall=free_price, tech=0, ingrs=0, energy=0}
			elseif regular_products[name] == nil then
				debug_print( "set ", name, " unknown_price")
				global.prices[name] = {overall=unknown_price, tech=0, ingrs=0, energy=0}
				global.unknowns[name] = true
			end
		end
	end

	regular_products = nil

	global.orig_resources = orig_resources
	global.new_resources = new_resources
	global.free_products = free_products
	
	global.prices_loop = 0
	global.ingredients_to_parse = ingredients_to_parse
end

--------------------------------------------------------------------------------------
local function update_products_recipe(recipe)
	local new_ingrs = {}
	local overall_price = 0
	local tech_price = 0
	local ingrs_price = 0
	local energy_price = recipe.energy * energy_cost
	
	local tech_name = global.recipes_tech[recipe.name]
	if tech_name then
		local tech_cost = global.techs_costs[tech_name]
		if tech_cost then
			tech_price = tech_cost * tech_amortization
		end
	end

	for _, ingr in pairs(recipe.ingredients) do
		local price_ingr = global.prices[ingr.name]
		if price_ingr == nil then
			return(nil)
		end
		ingrs_price = ingrs_price + ingr.amount * price_ingr.overall
	end
		
	local nb_prods_tot = 0
	for _, prod in pairs(recipe.products) do
		if prod.amount then
			nb_prods_tot = nb_prods_tot + prod.amount
		elseif prod.amount_min and prod.amount_max and prod.probability then
			nb_prods_tot = nb_prods_tot + (prod.amount_min + prod.amount_max) /2 * prod.probability
		else
			debug_print( "     NO amount : ", recipe.name, " ", prod.name )
		end
	end
		
	overall_price = (tech_price + ingrs_price + energy_price) * (1+commercial_margin)
	
	-- if nb_prods_tot == 0 then
		-- debug_print( "  nb_prods_tot 0 : ", recipe.name )
		-- for _, prod in pairs(recipe.products) do
			-- debug_print( "    prod = ", prod.name )
			-- debug_print( "		amount", prod.amount )
			-- debug_print( "		amount_min", prod.amount_min )
			-- debug_print( "		amount_max", prod.amount_max )
			-- debug_print( "		probability", prod.probability )
		-- end
	-- end
	
	debug_print("     OK nb prods = ", #recipe.products)
		
	for _, prod in pairs(recipe.products) do
		local p = global.prices[prod.name]
		if p == nil and (prod.amount or (prod.amount_min and prod.amount_max and prod.probability)) then
			if nb_prods_tot == 0 then
				p = unknown_price -- sometimes, probability can be 0, leading to nb_prods_tot=0 !
				global.prices[prod.name] = {
					overall = unknown_price, 
					tech = 0,
					ingrs = 0,
					energy = 0
				}
			else
				p = math.floor(overall_price / nb_prods_tot+0.5)
				global.prices[prod.name] = {
					overall = p, -- we want the price for 1 object !
					tech = math.floor(tech_price / nb_prods_tot+0.5),
					ingrs = math.floor(ingrs_price / nb_prods_tot+0.5),
					energy = math.floor(energy_price / nb_prods_tot+0.5)
				}
			end
			debug_print("     PR price ", prod.name, " = ", p)
			global.recipes_used[prod.name] = recipe.name
			new_ingrs[prod.name] = true -- mark this product as a new potential ingredient of further recipes
		else
			debug_print("     -- exist ", prod.name, " = ", p.overall)
		end
	end

	global.recipes_parsed[recipe.name] = true

	return(new_ingrs)
end

--------------------------------------------------------------------------------------
local function update_objects_prices_loop()
	local new_price = true
	local ingredients_to_parse2 = {}
	
	new_price = false
	global.prices_loop = global.prices_loop + 1
	
	debug_print("--------------------------------------------------------------------------------------")
	debug_print( "PASS ", global.prices_loop )
	
	for name_recipe, recipe in pairs(game.forces.player.recipes) do
		-- if not recipe.hidden and global.recipes_parsed[name_recipe] == nil then
		if global.recipes_parsed[name_recipe] == nil and 
			not(
				string.sub(name_recipe,1,3) == "rf-" -- reverse factory mod
				or  string.sub(name_recipe,1,5) == "repl-" -- dark matter mod
				or  string.sub(name_recipe,1,11) == "dry411srev-" -- z recycle mod
				-- or string.sub(name_recipe,1,4) == "box-" -- boxing mod
				-- or  string.sub(name_recipe,1,6) == "unbox-" -- boxing mod
				-- or  string.sub(name_recipe,1,6) == "boxed-" -- boxing mod
				-- or  string.sub(name_recipe,1,5) == "fill-" -- omnibarrel mod
				-- or  string.sub(name_recipe,1,6) == "empty-" -- omnibarrel mod
			)
		then
			for _, ingr in pairs(recipe.ingredients) do
				if global.ingredients_to_parse[ingr.name] == true then
					ingredients_to_parse3 = update_products_recipe(recipe)
					
					debug_print( "ingr    ", ingr.name, " -> recipe ", name_recipe )
					
					if ingredients_to_parse3 == nil then
						-- ingredient still not solved, rescan in next pass
						ingredients_to_parse2[ingr.name] = true 
					else
						-- ingredient solved, recipe products became new ingredients for next pass
						new_price = true
						for name_ingr, _ in pairs(ingredients_to_parse3) do
							ingredients_to_parse2[name_ingr] = true
						end
					end
				end
			end
		end
	end
	
	global.ingredients_to_parse = ingredients_to_parse2
	
	return(new_price)
end

--------------------------------------------------------------------------------------
local function update_objects_prices_end()
	debug_print( "nb loops = ", global.prices_loop)

	-- mark left unpriced potential ingredients as unknown
	-- no ! (ingredients left in this list are priced but they lead to recipes with ever-missing ingredients...
	
	-- for name_ingr, _ in pairs(global.ingredients_to_parse) do
		-- global.prices[name_ingr] = {overall=unknown_price, tech=0, ingrs=0, energy=0}
		-- global.unknowns[name_ingr] = true
	-- end
	
	-- mark all other left unpriced items and fluids

	for name_object, object in pairs(game.item_prototypes) do
		if global.prices[name_object] == nil then
			global.prices[name_object] = {overall=unknown_price, tech=0, ingrs=0, energy=0}
			global.unknowns[name_object] = true
		end
	end
	
	for name_object, object in pairs(game.fluid_prototypes) do
		if global.prices[name_object] == nil then
			global.prices[name_object] = {overall=unknown_price, tech=0, ingrs=0, energy=0}
			global.unknowns[name_object] = true
		end
	end
	
	global.ingredients_to_parse = nil
end

--------------------------------------------------------------------------------------
local function list_prices()
	debug_print("--------------------------------------------------------------------------------------")
	debug_print("list_prices")
	
	for name, price in pairs(global.prices) do
		local recipe_name = global.recipes_used[name]
		if recipe_name then
			debug_print("price ", name, "=", price.overall, "=", price.tech, "+", price.ingrs, "+", price.energy, " recipe=", recipe_name)
		else
			debug_print("price ", name, "=", price.overall, "=", price.tech, "+", price.ingrs, "+", price.energy, " recipe=NONE")
		end
	end

	debug_print("--------------------------------------------------------------------------------------")
	
	for name, object in pairs(game.item_prototypes) do
		local price = global.prices[name]
		if price == nil then
			debug_print("item ", name, "=NONE")
		else
			debug_print("item ", name, "=", price.overall, "=", price.tech, "+", price.ingrs, "+", price.energy)
		end
	end
	
	for name, object in pairs(game.fluid_prototypes) do
		local price = global.prices[name]
		if price == nil then price = "NONE" end
		if price == nil then
			debug_print("fluid ", name, "=NONE")
		else
			debug_print("fluid ", name, "=", price.overall, "=", price.tech, "+", price.ingrs, "+", price.energy)
		end
	end
end

--------------------------------------------------------------------------------------
local function export_prices()
	-- debug_print("export_prices")
	
	game.remove_path(prices_file)

	local s = "object;total price;tech cost;ingredients cost;energy cost;recipe;techno" .. "\n"
	
	if pcall(game.write_file,prices_file,s,true) then
		for name, price in pairs(global.prices) do
			local recipe_name = global.recipes_used[name]
			if recipe_name then
				local tech_name = global.recipes_tech[recipe_name]
				if tech_name == nil then tech_name = "" end
				s = name .. ";" .. price.overall .. ";" .. price.tech .. ";" .. price.ingrs .. ";" .. price.energy .. ";" .. recipe_name .. ";" .. tech_name .. "\n"
			else
				s = name .. ";" .. price.overall .. ";" .. price.tech .. ";" .. price.ingrs .. ";" .. price.energy .. ";" .. "" .. ";" .. "" .. "\n"
			end
			game.write_file(prices_file,s,true)
		end
	end
end

--------------------------------------------------------------------------------------
local function export_specials()
	-- debug_print("export_specials")
	
	game.remove_path(specials_file)
	
	local s = "-- Special items/fluids : resources/new resources/free price/unknown. If incorrect, add manually to config.lua and/or send to binbinhfr on forum : https://forums.factorio.com/viewtopic.php?f=92&t=30889\n"

	if pcall(game.write_file,specials_file,s,true) then
		game.write_file(specials_file,"	-- Original declared resources:\n",true)
		for name, _ in pairs(global.orig_resources) do
			s = "	[\"" .. name .. "\"] = resource_price,\n"
			game.write_file(specials_file,s,true)
		end
		game.write_file(specials_file,"\n",true)
		
		game.write_file(specials_file,"	-- Potential undeclared resources:\n",true)
		for name, _ in pairs(global.new_resources) do
			s = "	[\"" .. name .. "\"] = resource_price_new,\n"
			game.write_file(specials_file,s,true)
		end
		game.write_file(specials_file,"\n",true)
		
		game.write_file(specials_file,"	-- Free items/fluids (recipe without ingredients):\n",true)
		for name, _ in pairs(global.free_products) do
			s = "	[\"" .. name .. "\"] = free_price,\n"
			game.write_file(specials_file,s,true)
		end
		game.write_file(specials_file,"\n",true)
		
		game.write_file(specials_file,"	-- Unknown items/fluids (no recipe or loop):\n",true)
		for name, _ in pairs(global.unknowns) do
			s = "	[\"" .. name .. "\"] = unknown_price,\n"
			game.write_file(specials_file,s,true)
		end		
	end
end

--------------------------------------------------------------------------------------
local function update_groups()
	-- debug_print("--------------------------------------------------------------------------------------")
	-- debug_print("update_groups")
	
	global.groups = {}
	local groups = global.groups
	
	for name, recipe in pairs(game.forces.player.recipes) do
		local group_name = recipe.group.name
		if groups[group_name] == nil then groups[group_name] = recipe.group end
	end
end

--------------------------------------------------------------------------------------
local function list_groups()
	debug_print("--------------------------------------------------------------------------------------")
	debug_print("list_groups")
	
	for name, group in pairs(global.groups) do
		debug_print("group:", name, " childs=", #group.subgroups)
		for _, subgroup in pairs(group.subgroups) do
			debug_print("-> subgroup:", subgroup.name)
		end
	end
end

--------------------------------------------------------------------------------------
local function list_recipes()
	debug_print("--------------------------------------------------------------------------------------")
	debug_print("list_recipes")
	
	for name, recipe in pairs(game.forces.player.recipes) do
		debug_print("recipe:", name, " hidden=", recipe.hidden)
	end
end

--------------------------------------------------------------------------------------
local function get_hour()
	-- refresh hour and detect hour change ; hour increases continuously, and does not reset at 24.
	local surf = game.surfaces.nauvis

	if surf.always_day ~= global.always_day then
		global.always_day = surf.always_day
		global.hour_prev = -1
	end
		
	if surf.always_day then
		global.hour = math.floor(game.tick / 1042) -- one day is 25000 ticks 25000/24 = 1042, game starts at noon
		if global.hour ~= global.hour_prev then
			if global.hour%24 == 0 then 
				global.day = global.day + 1
				global.day_changed = true
			end
			global.hour_changed = 3
			global.hour_prev = global.hour
		end
	else
		local hour = math.floor(surf.daytime * 24) -- daytime [0,1], noon = 0.0, midnight = 0.5
		if hour ~= global.hour_prev then
			if hour == 0 then 
				global.day = global.day + 1
				global.day_changed = true
			end
			global.hour = 24*global.day + hour
			global.hour_changed = 3
			global.hour_prev = hour
		end
	end
end

--------------------------------------------------------------------------------------
local function compute_force_data(force_mem)
	local tot_taxes = force_mem.sales_taxes + force_mem.purchases_taxes
	local tot = force_mem.sales + force_mem.purchases + tot_taxes
	if tot == 0 then
		force_mem.tax_rate = 0
	else
		force_mem.tax_rate = math.floor(0.5+tot_taxes * 100 / tot)
	end
end

--------------------------------------------------------------------------------------
local function init_chest( chest )
	chest.auto = false
	chest.n_period=default_n_period
	chest.period=periods[default_n_period]
	
	chest.evaluation = 0 -- gross value of chest content
	
	chest.money_tot = 0 -- total sales or incomes
	chest.taxes_tot = 0 -- total taxes
	chest.tax_rate_tot = 0 -- total average tax rate
	
	chest.hour = global.hour -- hour of counter reset (initialized with last hour tick)
	chest.dhour = 0 -- hours since last reset
	chest.money_reset = 0 -- money at counter reset
	chest.taxes_reset = 0 -- taxes at counter reset
	chest.money = 0 -- sales or incomes since counter reset
	chest.taxes = 0 -- taxes or incomes since counter reset
	chest.tax_rate = 0 -- average tax rate
	chest.money_average = 0 -- money per day since counter reset
	
	chest.hour_period = global.hour -- starting hour of last period (used by counter reset)
	chest.money_tot_start_period = 0 -- money at beginning of last period
	chest.money_period = 0 -- money on last period
	
	chest.orders_tot = 0 -- total of the purchase list, without taxes
	chest.orders = {} -- purchase orders, list of {name, count}
	
	chest.editer = nil -- player who is currently editing
end

--------------------------------------------------------------------------------------
local function evaluate_chest(chest)
	local inv = chest.entity.get_inventory(defines.inventory.chest)
	local contents = inv.get_contents()
	local money = 0
	
	for name, count in pairs(inv.get_contents()) do
		local price = global.prices[name]
		if price ~= nil and price ~= 0 then
			money = money + count * price.overall
		end
	end

	chest.evaluation = money

	return(money)
end

--------------------------------------------------------------------------------------
local function sell_chest(chest,force_mem,tax_rate)
	if tax_rate == nil then
		tax_rate = global.tax_rates[chest.period]
	end
	local inv = chest.entity.get_inventory(defines.inventory.chest)
	local contents = inv.get_contents()
	local money1, tax1
	local money = 0
	local taxes = 0
	
	for name, count in pairs(inv.get_contents()) do
		local price = global.prices[name]
		if price ~= nil and price ~= 0 then
			money1 = count * price.overall
			if name == "ucoin" then
				tax1 = 0
			else
				tax1 = money1 * tax_rate / 100
			end
			money = money + money1 - tax1
			taxes = taxes + tax1
			
			inv.remove({name=name,count=count})
			
			if force_mem.transactions[name] == nil then
				force_mem.transactions[name] = 0
			end
			force_mem.transactions[name] = force_mem.transactions[name] - count
		end
	end

	chest.money_tot = chest.money_tot + money
	chest.taxes_tot = chest.taxes_tot + taxes
	
	force_mem.credits = force_mem.credits + money
	force_mem.sales = force_mem.sales + money
	force_mem.sales_taxes = force_mem.sales_taxes + taxes

	return(money)
end

--------------------------------------------------------------------------------------
local function buy_chest(chest,force_mem,tax_rate)
	if tax_rate == nil then
		tax_rate = global.tax_rates[chest.period]
	end
	local inv = chest.entity.get_inventory(defines.inventory.chest)
	local money1
	local tax1
	local money = 0
	local taxes = 0
	
	-- for i, order in pairs(chest.orders) do
	for i=#chest.orders,1,-1 do
		local order = chest.orders[i]
		local name = order.name
		-- if game.item_prototypes[name] == nil then
			-- table.remove(chest.orders,i)
		-- else
			if order.count > 0 then
				money1 = order.count * order.price
				if name == "ucoin" then
					tax1 = 0
				else
					tax1 = money1 * tax_rate / 100
				end
				if money1+tax1 <= force_mem.credits then
					-- can buy !
					local purchased = inv.insert({name=order.name,count=order.count})
					if purchased < order.count then
						money1 = purchased * order.price
						if name == "ucoin" then
							tax1 = 0
						else
							tax1 = money1 * tax_rate / 100
						end
					end
					force_mem.credits = force_mem.credits - money1 - tax1
					money = money + money1
					taxes = taxes + tax1
					if force_mem.transactions[name] == nil then
						force_mem.transactions[name] = 0
					end
					force_mem.transactions[name] = force_mem.transactions[name] + purchased
				else
					break
				end
			end
		-- end
	end

	chest.money_tot = chest.money_tot + money
	chest.taxes_tot = chest.taxes_tot + taxes
	
	force_mem.purchases = force_mem.purchases + money
	force_mem.purchases_taxes = force_mem.purchases_taxes + taxes

	return(money)
end

--------------------------------------------------------------------------------------
local function find_chest_sell(force_mem,ent)
	for _, chest in pairs(force_mem.chests_sell) do
		if chest.entity == ent then
			 return(chest)
		end
	end
	
	return(nil)
end

--------------------------------------------------------------------------------------
local function find_chest_buy(force_mem,ent)
	for _, chest in pairs(force_mem.chests_buy) do
		if chest.entity == ent then
			 return(chest)
		end
	end
	
	return(nil)
end

--------------------------------------------------------------------------------------
local function compute_chest_data(chest)
	local tot 
	
	if chest.sell_or_buy then
		tot = chest.money_tot + chest.taxes_tot 
	else
		tot = chest.money_tot 
	end
	if tot == 0 then
		chest.tax_rate_tot = 0
	else
		chest.tax_rate_tot = math.floor(0.5+100 * chest.taxes_tot  / tot)
	end

	chest.money = chest.money_tot - chest.money_reset
	chest.taxes = chest.taxes_tot - chest.taxes_reset
	
	chest.dhour = global.hour - chest.hour
	if chest.dhour == 0 then
		chest.money_average = 0
	else
		chest.money_average = math.floor(0.5+chest.money * 24 / chest.dhour)
	end
	
	if chest.sell_or_buy then
		tot = chest.money + chest.taxes
	else
		tot = chest.money
	end
	if tot == 0 then
		chest.tax_rate = 0
	else
		chest.tax_rate = math.floor(0.5+100 * chest.taxes  / tot)
	end

	if chest.sell_or_buy then
	else
		tot = 0
		for _, order in pairs(chest.orders) do
			tot = tot + order.count * order.price
		end
		
		chest.orders_tot = tot
		-- debug_print(tot)
	end
end

--------------------------------------------------------------------------------------
local function clean_orders()
	for name, force in pairs(game.forces) do
		local force_mem = global.force_mem[name]
		for i=#force_mem.chests_buy,1,-1 do
			local chest = force_mem.chests_buy[i]
			for i=#chest.orders,1,-1 do
				local order = chest.orders[i]
				if game.item_prototypes[order.name] == nil then
					table.remove(chest.orders,i)
				else
					local p = global.prices[order.name]
					if p then
						order.price = p.overall
					else
						order.price = 0
					end
				end
			end
		end
	end
end

--------------------------------------------------------------------------------------
local function clean_gui(gui)
	for _, guiname in pairs( gui.children_names ) do
		gui[guiname].destroy()
	end
end

--------------------------------------------------------------------------------------
local function build_bar( player )
	local gui1 = player.gui.top.flw_blkmkt
	
	if gui1 == nil then
		local player_mem = global.player_mem[player.index]
		-- debug_print("create gui player" .. player.name)
		gui1 = player.gui.top.add({type = "flow", name = "flw_blkmkt", direction = "horizontal", style = "flow_blkmkt_style"})
		gui1.add({type = "sprite-button", name = "but_blkmkt_main", sprite = "sprite_main_blkmkt", style = "sprite_main_blkmkt_style"})				
		player_mem.but_blkmkt_credits = gui1.add({type = "button", name = "but_blkmkt_credits", style = "button_blkmkt_credits_style"})				
	end
end

--------------------------------------------------------------------------------------
local function update_bar( player )
	local player_mem = global.player_mem[player.index]
	if player_mem.cursor_name == nil then
		local force_mem = global.force_mem[player.force.name]
		player_mem.but_blkmkt_credits.caption = format_money(force_mem.credits)
	else
		local price = global.prices[player_mem.cursor_name]
		if price == nil then 
			player_mem.but_blkmkt_credits.caption = "=NONE"
		else
			player_mem.but_blkmkt_credits.caption = "=" .. format_money(price.overall) 
		end
	end
end

--------------------------------------------------------------------------------------
local function update_bars( force )
	for _, player in pairs(force.players) do
		if player.connected then
			update_bar(player)
		end
	end
end

--------------------------------------------------------------------------------------
local function build_menu_gen( player, player_mem, open_or_close )
	if open_or_close == nil then
		open_or_close = (player.gui.left.frm_blkmkt_gen == nil)
	end
	
	if player.gui.left.frm_blkmkt_gen then
		player.gui.left.frm_blkmkt_gen.destroy()
	end
	
	if open_or_close then
		local gui1, gui2, gui3
		gui1 = player.gui.left.add({type = "frame", name = "frm_blkmkt_gen", caption = {"blkmkt-gui-blkmkt"}, style = "frame_blkmkt_style"})
		gui1 = gui1.add({type = "flow", name = "flw_blkmkt_gen", direction = "vertical", style = "flow_blkmkt_style"})
		
		gui2 = gui1.add({type = "flow", direction = "horizontal", style = "flow_blkmkt_style"})
		gui2.add({type = "label", caption = {"blkmkt-gui-gen-prices"}, style = "label_blkmkt_style"})
		gui2.add({type = "button", name = "but_blkmkt_gen_show_prices", caption = {"blkmkt-gui-gen-show-prices"}, 
			tooltip = {"blkmkt-gui-gen-show-prices-tt"}, style = "button_blkmkt_style"})
		gui2.add({type = "button", name = "but_blkmkt_gen_export_prices", caption = {"blkmkt-gui-gen-export-prices"}, 
			tooltip = {"blkmkt-gui-gen-export-prices-tt"}, style = "button_blkmkt_style"})
		gui2.add({type = "button", name = "but_blkmkt_gen_rescan_prices", caption = {"blkmkt-gui-gen-rescan-prices"}, 
			tooltip = {"blkmkt-gui-gen-rescan-prices-tt"}, style = "button_blkmkt_style"})
			
		gui2 = gui1.add({type = "flow", direction = "horizontal", style = "flow_blkmkt_style"})
		player_mem.chk_blkmkt_gen_pause = gui2.add({type = "checkbox", name = "chk_blkmkt_gen_pause", caption = {"blkmkt-gui-gen-pause"}, state = false, 
			tooltip = {"blkmkt-gui-gen-pause-tt"}, style = "checkbox_blkmkt_style"})
		gui2.add({type = "button", name = "but_blkmkt_gen_auto_all", caption = {"blkmkt-gui-gen-auto-all"}, 
			tooltip = {"blkmkt-gui-gen-auto-all-tt"}, style = "button_blkmkt_style"})
		gui2.add({type = "button", name = "but_blkmkt_gen_auto_none", caption = {"blkmkt-gui-gen-auto-none"}, 
			tooltip = {"blkmkt-gui-gen-auto-none-tt"}, style = "button_blkmkt_style"})
			
		gui2 = gui1.add({type = "flow", direction = "horizontal", style = "flow_blkmkt_style"})
		gui2.add({type = "button", name = "but_blkmkt_gen_sell_now", caption = {"blkmkt-gui-gen-sell-now"}, 
			tooltip = {"blkmkt-gui-gen-sell-now-tt",global.tax_rates[0]}, style = "button_blkmkt_style"})
		gui2.add({type = "button", name = "but_blkmkt_gen_buy_now", caption = {"blkmkt-gui-gen-buy-now"}, 
			tooltip = {"blkmkt-gui-gen-buy-now-tt",global.tax_rates[0]}, style = "button_blkmkt_style"})
			
		gui2 = gui1.add({type = "flow", direction = "horizontal", style = "flow_blkmkt_style"})
		gui2.add({type = "button", name = "but_blkmkt_gen_period_down", caption = "<", style = "button_blkmkt_style"})
		gui2.add({type = "button", name = "but_blkmkt_gen_period_up", caption = ">", style = "button_blkmkt_style"})
		player_mem.lbl_blkmkt_gen_period = gui2.add({type = "label", name = "lbl_blkmkt_gen_period", caption = {"blkmkt-gui-gen-period", 0,0}, 
			tooltip = {"blkmkt-gui-gen-period-tt"}, style = "label_blkmkt_style"})
		gui2.add({type = "button", name = "but_blkmkt_gen_period_set", caption = {"blkmkt-gui-gen-period-set"}, 
			tooltip = {"blkmkt-gui-gen-period-set-tt"}, style = "button_blkmkt_style"})

		gui2 = gui1.add({type = "table", colspan = 2, style = "table_blkmkt_style"})
		
		gui2.add({type = "label", caption = {"blkmkt-gui-gen-credits"}, 
			tooltip = {"blkmkt-gui-gen-credits-tt"}, style = "label_blkmkt_style"})
		player_mem.lbl_blkmkt_gen_credits = gui2.add({type = "label", name = "lbl_blkmkt_gen_credits", 
			tooltip = {"blkmkt-gui-gen-credits-tt"}, style = "label_blkmkt_style"})
		
		gui2.add({type = "label", caption = {"blkmkt-gui-gen-sales"}, 
			tooltip = {"blkmkt-gui-gen-sales-tt"}, style = "label_blkmkt_style"})
		player_mem.lbl_blkmkt_gen_sales = gui2.add({type = "label", name = "lbl_blkmkt_gen_sales", 
			tooltip = {"blkmkt-gui-gen-sales-tt"}, style = "label_blkmkt_style"})
		
		gui2.add({type = "label", caption = {"blkmkt-gui-gen-sales-taxes"}, 
			tooltip = {"blkmkt-gui-gen-sales-taxes-tt"}, style = "label_blkmkt_style"})
		player_mem.lbl_blkmkt_gen_sales_taxes = gui2.add({type = "label", name = "lbl_blkmkt_gen_sales_taxes", 
			tooltip = {"blkmkt-gui-gen-sales-taxes-tt"}, style = "label_blkmkt_style"})
		
		gui2.add({type = "label", caption = {"blkmkt-gui-gen-purchases"}, 
			tooltip = {"blkmkt-gui-gen-purchases-tt"}, style = "label_blkmkt_style"})
		player_mem.lbl_blkmkt_gen_purchases = gui2.add({type = "label", name = "lbl_blkmkt_gen_purchases", 
			tooltip = {"blkmkt-gui-gen-purchases-tt"}, style = "label_blkmkt_style"})
		
		gui2.add({type = "label", caption = {"blkmkt-gui-gen-purchases-taxes"}, 
			tooltip = {"blkmkt-gui-gen-purchases-taxes-tt"}, style = "label_blkmkt_style"})
		player_mem.lbl_blkmkt_gen_purchases_taxes = gui2.add({type = "label", name = "lbl_blkmkt_gen_purchases_taxes", 
			tooltip = {"blkmkt-gui-gen-purchases-taxes-tt"}, style = "label_blkmkt_style"})
		
		gui2.add({type = "label", caption = {"blkmkt-gui-gen-average-taxes"}, 
			tooltip = {"blkmkt-gui-gen-average-taxes-tt"}, style = "label_blkmkt_style"})
		player_mem.lbl_blkmkt_gen_tax_rate = gui2.add({type = "label", name = "lbl_blkmkt_gen_tax_rate", 
			tooltip = {"blkmkt-gui-gen-average-taxes-tt"}, style = "label_blkmkt_style"})
		
		gui2.add({type = "label", caption = {"blkmkt-gui-gen-credits-lastday"}, 
			tooltip = {"blkmkt-gui-gen-credits-lastday-tt"}, style = "label_blkmkt_style"})
		player_mem.lbl_blkmkt_gen_credits_lastday = gui2.add({type = "label", name = "lbl_blkmkt_gen_credits_lastday", 
			tooltip = {"blkmkt-gui-gen-credits-lastday-tt"}, style = "label_blkmkt_style"})
		
		gui1.add({type = "label", caption = {"blkmkt-gui-gen-transactions"}, 
			tooltip = {"blkmkt-gui-gen-transactions-tt"}, style = "label_blkmkt_style"})
		gui2 = gui1.add({type = "flow", name = "flw_blkmkt_gen_trans", direction = "horizontal", style = "flow_blkmkt_style"})
		gui2 = gui2.add({type = "scroll-pane", name = "scr_blkmkt_gen_trans", vertical_scroll_policy = "auto"})
		gui2.style.maximal_height = 150
		player_mem.scr_blkmkt_gen_trans = gui2
		
		gui1.add({type = "button", name = "but_blkmkt_gen_close", caption = {"blkmkt-gui-close"}, style = "button_blkmkt_style"})
	end
end

--------------------------------------------------------------------------------------
local function update_menu_gen( player, player_mem )
	if player.gui.left.frm_blkmkt_gen == nil then
		return
	end
	
	local force_mem = global.force_mem[player.force.name]
	player_mem.chk_blkmkt_gen_pause.state = force_mem.pause
	player_mem.lbl_blkmkt_gen_period.caption = {"blkmkt-gui-gen-period", force_mem.period,global.tax_rates[force_mem.period]}
	
	player_mem.lbl_blkmkt_gen_credits.caption = format_money(force_mem.credits)
	player_mem.lbl_blkmkt_gen_sales.caption = format_money(force_mem.sales)
	player_mem.lbl_blkmkt_gen_sales_taxes.caption = format_money(force_mem.sales_taxes)
	player_mem.lbl_blkmkt_gen_purchases.caption = format_money(force_mem.purchases)
	player_mem.lbl_blkmkt_gen_purchases_taxes.caption = format_money(force_mem.purchases_taxes)
	player_mem.lbl_blkmkt_gen_tax_rate.caption = force_mem.tax_rate .. "%"
	player_mem.lbl_blkmkt_gen_credits_lastday.caption = format_money(force_mem.credits_lastday) .. string.format(" (%+d%%)", force_mem.var_lastday)
	
	clean_gui(player_mem.scr_blkmkt_gen_trans)
	local gui2 = player_mem.scr_blkmkt_gen_trans.add({type = "table", name = "tab_blkmkt_gen_trans", colspan = 2, style = "table_blkmkt_style"})
	
	gui2.add({type = "label", caption = "name", style = "label_blkmkt_style"})
	gui2.add({type = "label", caption = "count", style = "label_blkmkt_style"})
	
	for name, transaction in pairs(force_mem.transactions) do
		-- debug_print(n, " ", name)
		gui2.add({type = "sprite-button", sprite = "item/" .. name, style = "sprite_obj_blkmkt_style"})
		gui2.add({type = "label", caption = transaction, style = "label_blkmkt_style"})
	end
end

--------------------------------------------------------------------------------------
local function build_menu_chest( player, player_mem, open_or_close )
	if open_or_close == nil then
		open_or_close = (player.gui.left.frm_blkmkt_chest == nil)
	end
	
	if player.gui.left.frm_blkmkt_chest then
		player.gui.left.frm_blkmkt_chest.destroy()
	end
	
	if open_or_close then
		local gui1, gui2, gui3
		gui1 = player.gui.left.add({type = "frame", name = "frm_blkmkt_chest", style = "frame_blkmkt_style"})
		player_mem.frm_blkmkt_chest = gui1
		gui1 = gui1.add({type = "flow", name = "flw_blkmkt_chest", direction = "vertical", style = "flow_blkmkt_style"})
		
		gui2 = gui1.add({type = "flow", direction = "horizontal", style = "flow_blkmkt_style"})
		player_mem.chk_blkmkt_chest_auto = gui2.add({type = "checkbox", name = "chk_blkmkt_chest_auto", caption = {"blkmkt-gui-chest-auto"}, state = false, 
			tooltip = {"blkmkt-gui-chest-auto-tt"}, style = "checkbox_blkmkt_style"})
		if player_mem.opened_chest.sell_or_buy then
			gui2.add({type = "button", name = "but_blkmkt_chest_now", caption = {"blkmkt-gui-chest-sell-now"}, 
				tooltip = {"blkmkt-gui-chest-sell-now-tt",global.tax_rates[0]}, style = "button_blkmkt_style"})
		else
			gui2.add({type = "button", name = "but_blkmkt_chest_now", caption = {"blkmkt-gui-chest-buy-now"}, 
				tooltip = {"blkmkt-gui-chest-buy-now-tt",global.tax_rates[0]}, style = "button_blkmkt_style"})
		end
		
		gui2 = gui1.add({type = "flow", direction = "horizontal", style = "flow_blkmkt_style"})
		gui2.add({type = "button", name = "but_blkmkt_chest_period_down", caption = "<", style = "button_blkmkt_style"})
		gui2.add({type = "button", name = "but_blkmkt_chest_period_up", caption = ">", style = "button_blkmkt_style"})
		player_mem.lbl_blkmkt_chest_period = gui2.add({type = "label", name = "lbl_blkmkt_chest_period", caption = {"blkmkt-gui-chest-period", 0,0}, 
			tooltip = {"blkmkt-gui-chest-period-tt"}, style = "label_blkmkt_style"})
		
		gui2 = gui1.add({type = "table", colspan = 2, style = "table_blkmkt_style"})
		
		player_mem.but_blkmkt_chest_evaluate =  gui2.add({type = "button", name = "but_blkmkt_chest_evaluate", caption = {"blkmkt-gui-chest-evaluate"}, 
			tooltip = {"blkmkt-gui-chest-evaluate-tt"}, style = "button_blkmkt_style"})
		player_mem.lbl_blkmkt_chest_evaluation = gui2.add({type = "label", name = "lbl_blkmkt_chest_evaluation", 
			tooltip = {"blkmkt-gui-chest-evaluation-tt"}, style = "label_blkmkt_style"})

		gui2.add({type = "label", caption = {"blkmkt-gui-chest-money-tot"}, 
			tooltip = {"blkmkt-gui-chest-money-tot-tt"}, style = "label_blkmkt_style"})
		player_mem.lbl_blkmkt_chest_money_tot = gui2.add({type = "label", name = "lbl_blkmkt_chest_money_tot", 
			tooltip = {"blkmkt-gui-chest-money-tot-tt"}, style = "label_blkmkt_style"})
		
		gui2.add({type = "label", caption = {"blkmkt-gui-chest-taxes-tot"}, 
			tooltip = {"blkmkt-gui-chest-taxes-tot-tt"}, style = "label_blkmkt_style"})
		player_mem.lbl_blkmkt_chest_taxes_tot = gui2.add({type = "label", name = "lbl_blkmkt_chest_taxes_tot", 
			tooltip = {"blkmkt-gui-chest-taxes-tot-tt"}, style = "label_blkmkt_style"})
		
		gui2.add({type = "label", caption = {"blkmkt-gui-chest-money-period"}, 
			tooltip = {"blkmkt-gui-chest-money-period-tt"}, style = "label_blkmkt_style"})
		player_mem.lbl_blkmkt_chest_money_period = gui2.add({type = "label", name = "lbl_blkmkt_chest_money_period", 
			tooltip = {"blkmkt-gui-chest-money-period-tt"}, style = "label_blkmkt_style"})
		
		gui2.add({type = "button", name = "but_blkmkt_chest_reset", caption = {"blkmkt-gui-chest-reset"}, style = "button_blkmkt_style"})
		player_mem.lbl_blkmkt_chest_dhour = gui2.add({type = "label", name = "lbl_blkmkt_chest_dhour", 
			tooltip = {"blkmkt-gui-chest-hours-tt"}, style = "label_blkmkt_style"})
		
		gui2.add({type = "label", caption = {"blkmkt-gui-chest-money"}, 
			tooltip = {"blkmkt-gui-chest-money-tt"}, style = "label_blkmkt_style"})
		player_mem.lbl_blkmkt_chest_money = gui2.add({type = "label", name = "lbl_blkmkt_chest_money", 
			tooltip = {"blkmkt-gui-chest-money-tt"}, style = "label_blkmkt_style"})
		
		gui2.add({type = "label", caption = {"blkmkt-gui-chest-taxes"}, 
			tooltip = {"blkmkt-gui-chest-taxes-tt"}, style = "label_blkmkt_style"})
		player_mem.lbl_blkmkt_chest_taxes = gui2.add({type = "label", name = "lbl_blkmkt_chest_taxes", 
			tooltip = {"blkmkt-gui-chest-taxes-tt"}, style = "label_blkmkt_style"})
		
		gui2.add({type = "label", caption = {"blkmkt-gui-chest-money-average"}, 
			tooltip = {"blkmkt-gui-chest-money-average-tt"}, style = "label_blkmkt_style"})
		player_mem.lbl_blkmkt_chest_money_average = gui2.add({type = "label", name = "lbl_blkmkt_chest_money_average", 
			tooltip = {"blkmkt-gui-chest-money-average-tt"}, style = "label_blkmkt_style"})

		if player_mem.opened_chest.sell_or_buy then
		else
			-- gui2 = gui1.add({type = "flow", direction = "horizontal", style = "flow_blkmkt_style"})
			gui2.add({type = "button", name = "but_blkmkt_chest_new", caption = {"blkmkt-gui-chest-new"}, style = "button_blkmkt_style"})
			gui2.add({type = "button", name = "but_blkmkt_chest_wipe", caption = {"blkmkt-gui-chest-wipe"}, style = "button_blkmkt_style"})
			
			gui2.add({type = "label", caption = {"blkmkt-gui-chest-orders"}, 
				tooltip = {"blkmkt-gui-chest-orders-tt"}, style = "label_blkmkt_style"})
			player_mem.lbl_blkmkt_chest_orders = gui2.add({type = "label", name = "lbl_blkmkt_chest_orders", 
				tooltip = {"blkmkt-gui-chest-orders-tt"}, style = "label_blkmkt_style"})
				
			gui2 = gui1.add({type = "flow", name = "flw_blkmkt_chest_orders", direction = "horizontal", style = "flow_blkmkt_style"})
			gui2 = gui2.add({type = "scroll-pane", name = "scr_blkmkt_chest_orders", vertical_scroll_policy = "auto"})
			gui2.style.maximal_height = 150
			player_mem.scr_blkmkt_chest_orders = gui2
		end
	end
end

--------------------------------------------------------------------------------------
local function update_menu_chest( player, player_mem, update_orders )
	if player.gui.left.frm_blkmkt_chest == nil then
		return
	end
	
	local chest = player_mem.opened_chest
	
	if chest.sell_or_buy then
		player_mem.frm_blkmkt_chest.caption = {"blkmkt-gui-chest-sell"}
	else
		player_mem.frm_blkmkt_chest.caption = {"blkmkt-gui-chest-buy"}
	end
	player_mem.chk_blkmkt_chest_auto.state = chest.auto
	player_mem.lbl_blkmkt_chest_period.caption = {"blkmkt-gui-chest-period", chest.period,global.tax_rates[chest.period]}
	player_mem.lbl_blkmkt_chest_evaluation.caption = format_money(chest.evaluation)
	
	player_mem.lbl_blkmkt_chest_money_tot.caption = format_money(chest.money_tot)
	player_mem.lbl_blkmkt_chest_taxes_tot.caption = format_money(chest.taxes_tot) .. " (" .. chest.tax_rate_tot .. "%)"
	player_mem.lbl_blkmkt_chest_money_period.caption = format_money(chest.money_period)
	player_mem.lbl_blkmkt_chest_dhour.caption = {"blkmkt-gui-chest-hours", chest.period*math.floor(chest.dhour/chest.period)}
	player_mem.lbl_blkmkt_chest_money.caption = format_money(chest.money)
	player_mem.lbl_blkmkt_chest_taxes.caption = format_money(chest.taxes) .. " (" .. chest.tax_rate .. "%)"
	player_mem.lbl_blkmkt_chest_money_average.caption = {"blkmkt-gui-chest-perday", format_money(chest.money_average)}
	
	if chest.sell_or_buy then
	else
		player_mem.lbl_blkmkt_chest_orders.caption = format_money(chest.orders_tot)
				
		if update_orders then
			clean_gui(player_mem.scr_blkmkt_chest_orders)
			local gui2 = player_mem.scr_blkmkt_chest_orders.add({type = "table", name = "tab_blkmkt_chest_orders", colspan = 4, style = "table_blkmkt_style"})
			
			gui2.add({type = "label"})
			gui2.add({type = "label", caption = "name", style = "label_blkmkt_style"})
			gui2.add({type = "label", caption = "count", style = "label_blkmkt_style"})
			gui2.add({type = "label", caption = "price", style = "label_blkmkt_style"})

			-- for i=#chest.orders,1,-1 do
				-- local order = chest.orders[i]
				-- if game.item_prototypes[order.name] == nil then
					-- table.remove(chest.orders,i)
				-- end
			-- end

			for n, order in pairs(chest.orders) do
				local name = order.name
				-- debug_print(n, " ", order.name)
				gui2.add({type = "button", name = "but_blkmkt_chd_" .. string.format("%2d",n) .. name, caption = "X", style = "button_blkmkt_style"})
				gui2.add({type = "sprite-button", name = "but_blkmkt_chi_" .. string.format("%2d",n) .. name, sprite = "item/" .. name, style = "sprite_obj_blkmkt_style"})
				gui2.add({type = "textfield", name = "txt_blkmkt_chc_" .. string.format("%2d",n) .. name, text = order.count, style = "textfield_blkmkt_style"})
				gui2.add({type = "label", caption = format_money(order.price), style = "label_blkmkt_style"})
			end
		end
	end
end

--------------------------------------------------------------------------------------
local function build_menu_objects( player, ask_sel, open_or_close )
	if open_or_close == nil then
		open_or_close = (player.gui.left.frm_blkmkt_itml == nil)
	end
	
	if player.gui.left.frm_blkmkt_itml then
		player.gui.left.frm_blkmkt_itml.destroy()
	end
	
	if open_or_close then
		local player_mem = global.player_mem[player.index]
		local gui1, gui2, gui3
		gui1 = player.gui.left.add({type = "frame", name = "frm_blkmkt_itml", caption = {"blkmkt-gui-objects-list"}, style = "frame_blkmkt_style"})
		gui1 = gui1.add({type = "flow", name = "flw_blkmkt_itml", direction = "vertical", style = "flow_blkmkt_style"})
		-- gui1.style.minimal_height = 500
		
		gui2 = gui1.add({type = "scroll-pane", name = "scr_blkmkt_itml", vertical_scroll_policy = "auto"}) -- , style = "scroll_pane_blkmkt_style"
		gui2.style.maximal_height = 450
		player_mem.scr_blkmkt_recl = gui2
		gui3 = gui2.add({type = "table", name = "tab_blkmkt_itml1", colspan = 6, style = "table_blkmkt_style"})
		
		local n = 0
		
		for name, group in pairs(global.groups) do
			if player_mem.group_sel_name == nil then
				player_mem.group_sel_name = name
			end
			gui3.add({type = "sprite-button", name = "but_blkmkt_ilg_" .. string.format("%2d",n) .. name, sprite = "item-group/" .. name, tooltip = group.name, style = "sprite_group_blkmkt_style"})
			n=n+1
		end
		
		gui3 = gui2.add({type = "table", name = "tab_blkmkt_itml2", colspan = 10, style = "table_blkmkt_style"})
		
		local group = global.groups[player_mem.group_sel_name]
		
		for name, object in pairs(game.item_prototypes) do
			if object.group == group and not object.has_flag("hidden") then
				local p = global.prices[name]
				if p ~= nil and p.overall ~= 0 then
					gui3.add({type = "sprite-button", name = "but_blkmkt_ili_" .. string.format("%2d",n) .. name, sprite = "item/" .. name, 
					tooltip = {"blkmkt-gui-tt-object-price",object.localised_name,format_money(p.overall)}, style = "sprite_obj_blkmkt_style"})
					n=n+1
				end
			end
		end
		
		-- for name, object in pairs(game.fluid_prototypes) do
			-- if object.group == group not object.has_flag("hidden") then
				-- local p = global.prices[name]
				-- if p ~= nil and p.overall ~= 0 then
					-- gui3.add({type = "sprite-button", name = "but_blkmkt_ili_" .. string.format("%2d",n) .. name, sprite = "fluid/" .. name, 
					-- tooltip = {"blkmkt-gui-tt-object-price",object.localised_name,format_money(p.overall)}, style = "sprite_obj_blkmkt_style"})
					-- n=n+1
				-- end
			-- end
		-- end
		
		if ask_sel then
			gui1.add({type = "button", name = "but_blkmkt_itml_cancel", caption = {"blkmkt-gui-cancel"}, style = "button_blkmkt_style"})
		else
			gui1.add({type = "button", name = "but_blkmkt_itml_close", caption = {"blkmkt-gui-close"}, style = "button_blkmkt_style"})
		end
		-- player_mem.chk_blkmkt_itml_close = gui1.add({type = "checkbox", name = "chk_blkmkt_itml_close", caption = {"blkmkt-gui-auto-close"}, state = player_mem.auto_close, style = "checkbox_blkmkt_style"})
	end
end

--------------------------------------------------------------------------------------
local function update_guis_force(force)
	for _, player in pairs(force.players) do
		if player.connected then
			local player_mem = global.player_mem[player.index]
			update_bar(player)
			update_menu_gen(player,player_mem)
			update_menu_chest(player,player_mem,false)
		end
	end
end

--------------------------------------------------------------------------------------
local function close_guis()
	for _, player in pairs(game.players) do
		if player.connected then
			build_menu_gen(player,player_mem,false)
			build_menu_chest(player,player_mem,false)
		end
	end
end

--------------------------------------------------------------------------------------
local function init_globals()
	-- initialize or update general globals of the mod
	debug_print( "init_globals" )
	
	global.tick = global.tick or 0
	global.force_mem = global.force_mem or {}
	global.player_mem = global.player_mem or {}
	
	global.always_day = game.surfaces.nauvis.always_day
	global.day = global.day or 0 -- day (changes at noon)
	global.day_changed = false
	global.hour = global.hour or -1 -- hour (always increases, does not reset at 24...)
	global.hour_prev = global.hour_prev or -1
	global.hour_changed = 0
	
	global.prices_loop = global.prices_loop or -1

	if global.groups == nil then
		update_groups()
	end
	
	if global.techs_costs == nil then -- costs for every tech
		update_techs_costs() 
	end
	
	global.orig_resources = global.orig_resources or {} -- items undeclared as resources
	global.new_resources = global.new_resources or {} -- items that could be undeclared resources (used as ingredients but never produced)
	global.free_products = global.free_products or {} -- items with no prices
	global.unknowns = global.unknowns or {} -- items with no prices
	
	if global.prices == nil then -- prices for every item/fluid
		update_objects_prices_start()
	end
	
	if global.ask_rescan == nil then global.ask_rescan = false end
	
	if global.tax_rates == nil then
		init_tax_rates()
	end
end

--------------------------------------------------------------------------------------
local function init_force(force)
	if global.force_mem == nil then return end
	
	-- initialize or update per force globals of the mod
	debug_print( "init_force ", force.name )
	
	global.force_mem[force.name] = global.force_mem[force.name] or {}
	local force_mem = global.force_mem[force.name]
	
	if force_mem.pause == nil then force_mem.pause = false end
	force_mem.n_period = force_mem.n_period or default_n_period
	force_mem.period = periods[default_n_period]
	-- force_mem.credits = force_mem.credits or ((debug_status == 1) and 1000000 or 0)
	force_mem.credits = force_mem.credits or 0
	force_mem.credits_startday = force_mem.credits_startday or global.day -- credits at beginning of last day
	force_mem.credits_lastday = force_mem.credits_lastday or 0 -- credits during last day
	force_mem.var_lastday = force_mem.var_lastday or 0 -- variation od credit during last day
	force_mem.sales = force_mem.sales or 0 -- sum of all net sales
	force_mem.sales_taxes = force_mem.sales_taxes or 0 -- sum of all sales taxes
	force_mem.purchases = force_mem.purchases or 0 -- sum of all net purchases
	force_mem.purchases_taxes = force_mem.purchases_taxes or 0 -- sum of all purchases taxes
	force_mem.tax_rate = force_mem.tax_rate or 0 -- average overall tax rate
	force_mem.transactions = force_mem.transactions or {} -- balance per object (in nb of objects)
	
	force_mem.chests_sell = force_mem.chests_sell or {}
	force_mem.chests_buy = force_mem.chests_buy or {}
end

--------------------------------------------------------------------------------------
local function init_forces()
	for _, force in pairs(game.forces) do
		init_force(force)
	end
	
	clean_orders()
end

--------------------------------------------------------------------------------------
local function init_player(player)
	if global.player_mem == nil then return end
	
	-- initialize or update per player globals of the mod, and reset the gui
	debug_print( "init_player ", player.name, " connected=", player.connected )
	
	global.player_mem[player.index] = global.player_mem[player.index] or {}
	
	local player_mem = global.player_mem[player.index]
	if player_mem.auto_close == nil then player_mem.auto_close = false end
	player_mem.group_sel_name = player_mem.group_sel_name or nil
	player_mem.object_sel_name = player_mem.object_sel_name or nil
	
	player_mem.opened = player_mem.opened or nil
	player_mem.opened_chest = player_mem.opened_chest or nil
	player_mem.order_sel_n = player_mem.order_sel_n or 0 -- currently edited order
	
	if player.connected then
		build_bar(player)
		update_bar(player)
	end
end

--------------------------------------------------------------------------------------
local function init_players()
	for _, player in pairs(game.players) do
		init_player(player)
	end
end

--------------------------------------------------------------------------------------
local function on_init() 
	-- called once, the first time the mod is loaded on a game (new or existing game)
	debug_print( "RAZ" )
	debug_print( "CLEAR" )
	debug_print( "on_init" )
	
	init_globals()
	init_forces()
	init_players()
end

script.on_init(on_init)

--------------------------------------------------------------------------------------
local function on_configuration_changed(data)
	-- detect any mod or game version change
	if data.mod_changes ~= nil then		
		local changes = data.mod_changes[debug_mod_name]
		if changes ~= nil then
			debug_print( "RAZ" )
			debug_print( "CLEAR" )
			debug_print( "update mod: ", debug_mod_name, " ", tostring(changes.old_version), " to ", tostring(changes.new_version) )
		
			init_globals()
			init_forces()
			init_players()
			
			-- if changes.old_version and older_version(changes.old_version, "1.0.0") then
				-- migrations
			-- end
		end
		
		-- global.ask_rescan = true
		close_guis()

		-- if any other mod install or uninstall, rescan prices ! and clean orders

		update_groups()

		if global.prices_loop == -1 then
			debug_print( "update other mods" )
			update_techs_costs()
			update_objects_prices_start()		
		end

		clean_orders()
	end
end

script.on_configuration_changed(on_configuration_changed)

--------------------------------------------------------------------------------------
local function on_force_created(event)
	-- called at player creation
	local force = event.force
	debug_print( "force created ", force.name )
	
	init_force(force)
end

script.on_event(defines.events.on_force_created, on_force_created )

--------------------------------------------------------------------------------------
local function on_forces_merging(event)
	local force1 = event.source
	local force2 = event.destination
	debug_print( "force merging ", force1.name, " into ", force2.name )
	
	local force_mem1 = global.force_mem[force1.name]
	local force_mem2 = global.force_mem[force2.name]
	
	force_mem2.credits = force_mem2.credits + force_mem1.credits
	force_mem2.credits_startday = force_mem2.credits_startday + force_mem1.credits_startday
	force_mem2.credits_lastday = force_mem2.credits_lastday + force_mem1.credits_lastday
	force_mem2.var_lastday = force_mem2.var_lastday + force_mem1.var_lastday
	force_mem2.sales = force_mem2.sales + force_mem1.sales
	force_mem2.sales_taxes = force_mem2.sales_taxes + force_mem1.sales_taxes
	force_mem2.purchases = force_mem2.purchases + force_mem1.purchases
	force_mem2.purchases_taxes = force_mem2.purchases_taxes + force_mem1.purchases_taxes
	
	concat_lists(force_mem2.chests_sell, force_mem1.chests_sell)
	concat_lists(force_mem2.chests_buy, force_mem1.chests_buy)

	-- global.force_mem[force1.name] = nil

	compute_force_data(force_mem2)
	update_guis_force(force2)
end

script.on_event(defines.events.on_forces_merging, on_forces_merging )

--------------------------------------------------------------------------------------
local function on_player_created(event)
	-- called at player creation
	local player = game.players[event.player_index]
	debug_print( "player created ", player.name )
	
	init_player(player)
	
	if debug_status == 1 then
		local inv = player.get_inventory(defines.inventory.player_quickbar)
		inv.insert({name="fast-inserter", count=50})
		inv.insert({name="trading-chest-sell", count=10})
		inv.insert({name="trading-chest-buy", count=10})
		inv.insert({name="ucoin", count=100000})
	end
end

script.on_event(defines.events.on_player_created, on_player_created )

--------------------------------------------------------------------------------------
local function on_player_joined_game(event)
	-- called in SP(once) and MP(at every connect), eventually after on_player_created
	local player = game.players[event.player_index]
	debug_print( "player joined ", player.name )
	
	init_player(player)
end

script.on_event(defines.events.on_player_joined_game, on_player_joined_game )

--------------------------------------------------------------------------------------
local function on_player_cursor_stack_changed(event)
	local player = game.players[event.player_index]
	local player_mem = global.player_mem[player.index]
	
	-- debug_print( "on_player_cursor_stack_changed ", player.name )
	
	if player.cursor_stack.valid_for_read then
		player_mem.cursor_name = player.cursor_stack.name
	else	
		player_mem.cursor_name = nil
	end
	
	update_bar(player)
end

script.on_event(defines.events.on_player_cursor_stack_changed, on_player_cursor_stack_changed )

--------------------------------------------------------------------------------------
local function on_creation( event )
	local ent = event.created_entity
	local ent_name = ent.name
	local sell_or_buy = nil
	
	-- debug_print( "creation ", ent_name )
	
	if ent_name == "trading-chest-sell" then
		sell_or_buy = true
	elseif ent_name == "trading-chest-buy" then
		sell_or_buy = false
	end
	
	if sell_or_buy ~= nil then
		local force_mem = global.force_mem[ent.force.name]
		local chest = {	entity=ent, sell_or_buy = sell_or_buy }
		init_chest(chest)

		if sell_or_buy then
			table.insert(force_mem.chests_sell,chest)
		else
			chest.orders = {
				{name = "coal", count = 200, price = global.prices["coal"].overall },
				{name = "copper-ore", count = 100, price = global.prices["copper-ore"].overall },
				{name = "iron-ore", count = 100, price = global.prices["iron-ore"].overall },
				{name = "copper-plate", count = 50, price = global.prices["copper-plate"].overall },
				{name = "iron-plate", count = 50, price = global.prices["iron-plate"].overall },
			}
			table.insert(force_mem.chests_buy,chest)
		end
		
		compute_chest_data(chest)
	end
end

script.on_event(defines.events.on_built_entity, on_creation )
script.on_event(defines.events.on_robot_built_entity, on_creation )

--------------------------------------------------------------------------------------
local function on_destruction( event )
	local ent = event.entity
	local ent_name = ent.name
	
	-- debug_print( "destruction ", ent_name )
	
	if ent_name == "trading-chest-sell" then
		local force_mem = global.force_mem[ent.force.name]
		for i, chest in pairs(force_mem.chests_sell) do
			if chest.entity == ent then
				table.remove(force_mem.chests_sell,i)
				break
			end
		end
		
	elseif ent_name == "trading-chest-buy" then
		local force_mem = global.force_mem[ent.force.name]
		for i, chest in pairs(force_mem.chests_buy) do
			if chest.entity == ent then
				table.remove(force_mem.chests_buy,i)
				break
			end
		end
	end
end

script.on_event(defines.events.on_entity_died, on_destruction )
script.on_event(defines.events.on_robot_pre_mined, on_destruction )
script.on_event(defines.events.on_preplayer_mined_item, on_destruction )

--------------------------------------------------------------------------------------
local function on_tick(event)
	if global.tick >= 99 then 
		global.tick = 0
		
		if global.ask_rescan then
			global.ask_rescan = false
			message_all({"blkmkt-gui-askrescan"})
		end
		
	elseif global.tick%20 == 1 then
		-- check hour change
		get_hour()

		if global.prices_loop > -1 then
			message_all({"blkmkt-gui-rescan-progr",global.prices_loop})
			
			if (not update_objects_prices_loop()) or global.prices_loop > 15 then
				update_objects_prices_end()
				export_specials()
				clean_orders()
				global.prices_loop = -1
				message_all({"blkmkt-gui-rescan-done"})
			end
		end
		
		-- manage opened chests
		for _, player in pairs(game.players) do
			if player.connected then
				local player_mem = global.player_mem[player.index]
				local opened = player.opened
				
				if opened then
					if opened ~= player_mem.opened then
						local force_mem = global.force_mem[player.force.name]
						
						if opened.name == "trading-chest-sell" then
							local chest = find_chest_sell(force_mem,opened)
							if chest then
								if chest.editer == nil or not chest.editer.connected then
									player_mem.opened_chest = chest
									chest.editer = player
									build_menu_chest(player,player_mem,true)
									update_menu_chest(player,player_mem,true)
								else
									player.print({"blkmkt-gui-chest-edited",chest.editer.name})
								end
							end
						elseif opened.name == "trading-chest-buy" then
							local chest = find_chest_buy(force_mem,opened)
							if chest then
								if chest.editer == nil or not chest.editer.connected then
									player_mem.opened_chest = chest
									chest.editer = player
									build_menu_chest(player,player_mem,true)
									update_menu_chest(player,player_mem,true)
								else
									player.print({"blkmkt-gui-chest-edited",chest.editer.name})
								end
							end
						end
						
						player_mem.opened = opened
					end
				else
					if player_mem.opened then
						if player_mem.opened_chest then
							player_mem.opened_chest.editer = nil
							
							build_menu_chest(player,player_mem,false)
						end
						
						player_mem.opened = nil
						player_mem.opened_chest = nil
					end
				end
			end
		end
	
	elseif global.tick == 38 then
		-- do sales
		
		if global.hour_changed == 3 then
			global.hour_changed = global.hour_changed - 1 
			
			for name, force in pairs(game.forces) do
				local force_mem = global.force_mem[name]
				local money = 0
				-- debug_print("force ", name, " chests=",#force_mem.chests_sell)
				for i=#force_mem.chests_sell,1,-1 do
					local chest = force_mem.chests_sell[i]
					if global.hour % chest.period == 0 then
						chest.hour_period = global.hour
						if chest.auto and not force_mem.pause then
							local ent = chest.entity
							if ent and ent.valid then
								-- debug_print("sell !" )
								money = money + sell_chest(chest,force_mem)
							else
								table.remove(force_mem.chests_sell,i)
							end
						end
					end
				end
				if money ~= 0 then
					update_bars(force)
				end
			end
		end
		
	elseif global.tick == 58 then
		-- do purchases
		
		if global.hour_changed == 2 then
			global.hour_changed = global.hour_changed - 1 
			
			for name, force in pairs(game.forces) do
				local force_mem = global.force_mem[name]
				local money = 0
				-- debug_print("force ", name, " chests=",#force_mem.chests_buy)
				for i=#force_mem.chests_buy,1,-1 do
					local chest = force_mem.chests_buy[i]
					if global.hour % chest.period == 0 then
						chest.hour_period = global.hour
						if chest.auto and not force_mem.pause then
							local ent = chest.entity
							if ent and ent.valid then
								-- debug_print("buy !" )
								money = money + buy_chest(chest,force_mem)
							else
								table.remove(force_mem.chests_buy,i)
							end
						end
					end
				end
				if money ~= 0 then
					update_bars(force)
				end
			end
		end
		
	elseif global.tick == 78 then
		if global.hour_changed == 1 then
			global.hour_changed = global.hour_changed - 1 
			
			-- compute period averages
		
			for name, force in pairs(game.forces) do
				local force_mem = global.force_mem[name]
				for i=#force_mem.chests_sell,1,-1 do
					local chest = force_mem.chests_sell[i]
					if chest.auto and (not force_mem.pause) and global.hour % chest.period == 0 then
						compute_chest_data(chest)
						chest.money_period = chest.money_tot - chest.money_tot_start_period
						-- debug_print(chest.money_tot_start_period, " -> ", chest.money_tot, " = ", chest.money_period)
						chest.money_tot_start_period = chest.money_tot
					end
				end
				for i=#force_mem.chests_buy,1,-1 do
					local chest = force_mem.chests_buy[i]
					if chest.auto and (not force_mem.pause) and global.hour % chest.period == 0 then
						compute_chest_data(chest)
						chest.money_period = chest.money_tot - chest.money_tot_start_period
						-- debug_print(chest.money_tot_start_period, " -> ", chest.money_tot, " = ", chest.money_period)
						chest.money_tot_start_period = chest.money_tot
					end
				end
				compute_force_data(force_mem)
			end
			
			-- update average display on opened chests
			
			for _, player in pairs(game.players) do
				if player.connected then
					local player_mem = global.player_mem[player.index]
					if player_mem.opened_chest then
						update_menu_chest(player,player_mem,false)
					end
					update_menu_gen(player,player_mem)
				end
			end
		end
		
	elseif global.tick == 98 then
		if global.day_changed then
			-- day change
			global.day_changed = false
			
			for name, force in pairs(game.forces) do
				local force_mem = global.force_mem[name]
				force_mem.credits_lastday = force_mem.credits - force_mem.credits_startday
				if force_mem.credits_startday == 0 then
					force_mem.var_lastday = 0
				else
					force_mem.var_lastday = math.floor(0.5+force_mem.credits_lastday * 100 / force_mem.credits_startday)
				end
				force_mem.credits_startday = force_mem.credits
			end
			
			for _, player in pairs(game.players) do
				if player.connected then
					local player_mem = global.player_mem[player.index]
					update_menu_gen(player,player_mem)
				end
			end
		end
	end
	
	global.tick = global.tick + 1 
end

script.on_event(defines.events.on_tick, on_tick)

--------------------------------------------------------------------------------------
local function on_gui_click(event)
	local player = game.players[event.player_index]
	local force = player.force
	local player_mem = global.player_mem[player.index]
	local event_name = event.element.name
	local prefix = string.sub(event_name,1,15)
	local nix = tonumber(string.sub(event_name,16,17))
	local suffix = string.sub(event_name,18)
	
	-- debug_print( "on_gui_click ", player.name, " ", event_name )

	if event_name == "but_blkmkt_main" then
		build_menu_gen(player,player_mem)
		update_menu_gen( player, player_mem )
		
	elseif event_name == "but_blkmkt_credits"then
		build_menu_objects(player,false)

	elseif event_name == "but_blkmkt_gen_show_prices"then
		build_menu_objects(player,false)
		
	elseif event_name == "but_blkmkt_gen_export_prices"then
		if debug_status == 1 then
			-- debug_print("RAZ")
			list_groups()
			list_techs_costs()
			list_prices()
			list_recipes()
		end
		
		export_prices()
		export_specials()
	
	elseif event_name == "but_blkmkt_gen_rescan_prices"then
		update_groups()
		update_techs_costs()
		update_objects_prices_start()
		
	elseif event_name == "chk_blkmkt_gen_pause" then
		local force_mem = global.force_mem[force.name]
		
		force_mem.pause = event.element.state
		
		update_guis_force(force)
		
	elseif event_name == "but_blkmkt_gen_auto_all"then
		local force_mem = global.force_mem[force.name]

		for _, chest in pairs(force_mem.chests_sell) do
			chest.auto = true
		end
	
		for _, chest in pairs(force_mem.chests_buy) do
			chest.auto = true
		end
	
		-- update_menu_chest(player,player_mem,false)
		update_guis_force(force)
		
	elseif event_name == "but_blkmkt_gen_auto_none"then
		local force_mem = global.force_mem[force.name]

		for _, chest in pairs(force_mem.chests_sell) do
			chest.auto = false
		end
	
		for _, chest in pairs(force_mem.chests_buy) do
			chest.auto = false
		end
	
		-- update_menu_chest(player,player_mem,false)
		update_guis_force(force)
		
	elseif event_name == "but_blkmkt_gen_sell_all"then
		local force_mem = global.force_mem[force.name]
		local money = 0
		
		for i=#force_mem.chests_sell,1,-1 do
			local chest = force_mem.chests_sell[i]
			local ent = chest.entity
			if ent and ent.valid then
				-- debug_print("sell !" )
				money = money + sell_chest(chest,force_mem)
			else
				table.remove(force_mem.chests_sell,i)
			end
		end
		if money ~= 0 then
			compute_force_data(force_mem)
			update_guis_force(force)
			-- update_bars(force)
			-- update_menu_gen(player,player_mem)
			-- update_menu_chest(player,player_mem,true)
		end
		
	elseif event_name == "but_blkmkt_gen_buy_all"then
		local force_mem = global.force_mem[force.name]
		local money = 0
		
		for i=#force_mem.chests_buy,1,-1 do
			local chest = force_mem.chests_buy[i]
			local ent = chest.entity
			if ent and ent.valid then
				-- debug_print("buy !" )
				money = money + buy_chest(chest,force_mem)
			else
				table.remove(force_mem.chests_buy,i)
			end
		end
		if money ~= 0 then
			compute_force_data(force_mem)
			update_guis_force(force)
			-- update_bars(force)
			-- update_menu_gen(player,player_mem)
			-- update_menu_chest(player,player_mem,true)
		end
		
	elseif event_name == "but_blkmkt_gen_period_down" then
		build_menu_objects(player,false,false)

		local force_mem = global.force_mem[force.name]

		if force_mem.n_period > 2 then
			force_mem.n_period = force_mem.n_period - 1
			force_mem.period = periods[force_mem.n_period]
		end
		
		-- update_menu_gen(player,player_mem)
		update_guis_force(force)
		
	elseif event_name == "but_blkmkt_gen_period_up" then
		build_menu_objects(player,false,false)

		local force_mem = global.force_mem[force.name]

		if force_mem.n_period < #periods then
			force_mem.n_period = force_mem.n_period + 1
			force_mem.period = periods[force_mem.n_period]
		end
		
		-- update_menu_gen(player,player_mem)
		update_guis_force(force)
		
	elseif event_name == "but_blkmkt_gen_period_set" then
		build_menu_objects(player,false,false)

		local force_mem = global.force_mem[force.name]

		for _, chest in pairs(force_mem.chests_sell) do
			chest.n_period = force_mem.n_period
			chest.period = force_mem.period
		end
		
		for _, chest in pairs(force_mem.chests_buy) do
			chest.n_period = force_mem.n_period
			chest.period = force_mem.period
		end

		-- update_menu_chest(player,player_mem,false)
		update_guis_force(force)

	elseif event_name == "but_blkmkt_gen_close" then
		build_menu_gen(player,player_mem,false)

	elseif prefix == "but_blkmkt_ilg_" then -- click group in item list
		-- debug_print(suffix)
		player_mem.group_sel_name = suffix
		build_menu_objects(player,false,true)

	elseif prefix == "but_blkmkt_ili_" then -- click item in item list
		-- debug_print(suffix)
		player_mem.object_sel_name = suffix

		local chest = player_mem.opened_chest
		
		if chest and player_mem.order_sel_n ~= 0 then
			build_menu_objects(player,false,false)
			
			local order = chest.orders[player_mem.order_sel_n]
			
			if order then
				order.name = player_mem.object_sel_name
				local p = global.prices[order.name]
				if p == nil then
					order.price = 0
				else
					order.price = p.overall
				end
			end
			
			compute_chest_data(chest)
			update_menu_chest(player,player_mem,true)
			player_mem.order_sel_n = 0
		end
		
	elseif event_name == "but_blkmkt_itml_close" then
		build_menu_objects(player,false,false)
		
	elseif event_name == "but_blkmkt_itml_cancel" then
		player_mem.order_sel_n = 0
		
		build_menu_objects(player,false,false)
		
	elseif event_name == "chk_blkmkt_chest_auto" then
		build_menu_objects(player,false,false)
		
		player_mem.opened_chest.auto = player_mem.chk_blkmkt_chest_auto.state
		
	elseif event_name == "but_blkmkt_chest_now" then -- sell or buy chest now
		build_menu_objects(player,false,false)
		
		local force_mem = global.force_mem[force.name]
		local chest = player_mem.opened_chest
		
		if chest.sell_or_buy then
			sell_chest(chest,force_mem,global.tax_rates[0])
		else
			buy_chest(chest,force_mem,global.tax_rates[0])
		end
		compute_chest_data(chest)
		update_menu_chest(player,player_mem,false)
		update_bars(force)
		
	elseif event_name == "but_blkmkt_chest_period_down" then
		build_menu_objects(player,false,false)
		
		local chest = player_mem.opened_chest
		
		if chest.n_period > 2 then
			chest.n_period = chest.n_period - 1
			chest.period = periods[chest.n_period]
		end
		
		update_menu_chest(player,player_mem,false)
		
	elseif event_name == "but_blkmkt_chest_period_up" then
		build_menu_objects(player,false,false)
		
		local chest = player_mem.opened_chest
		
		if chest.n_period < #periods then
			chest.n_period = chest.n_period + 1
			chest.period = periods[chest.n_period]
		end
		
		update_menu_chest(player,player_mem,false)
		
	elseif event_name == "but_blkmkt_chest_evaluate" then
		build_menu_objects(player,false,false)
		
		evaluate_chest(player_mem.opened_chest)
		update_menu_chest(player,player_mem,false)
		
	elseif event_name == "but_blkmkt_chest_reset" then
		build_menu_objects(player,false,false)
		
		local chest = player_mem.opened_chest
		
		chest.hour = chest.hour_period
		chest.money_reset = chest.money_tot
		chest.taxes_reset = chest.taxes_tot
		chest.money = 0
		chest.taxes = 0
		chest.money_average = 0
		
		compute_chest_data(chest)
		update_menu_chest(player,player_mem,false)

	elseif event_name == "but_blkmkt_chest_new" then -- new order
		build_menu_objects(player,false,false)
		
		local chest = player_mem.opened_chest
		
		table.insert(chest.orders,1,{name="coal", count=0, price=global.prices.coal.overall})
		update_menu_chest(player,player_mem,true)
	
	elseif event_name == "but_blkmkt_chest_wipe" then -- wipe orders
		build_menu_objects(player,false,false)
		
		local chest = player_mem.opened_chest
		
		chest.orders = {}
		compute_chest_data(chest)
		update_menu_chest(player,player_mem,true)
	
	elseif prefix == "but_blkmkt_chd_" then -- del item in orders list
		build_menu_objects(player,false,false)
		
		local chest = player_mem.opened_chest
		-- debug_print(nix, " ", suffix)
		
		table.remove(chest.orders,nix)
		compute_chest_data(chest)
		update_menu_chest(player,player_mem,true)

	elseif prefix == "but_blkmkt_chi_" then -- change order's item
		-- debug_print(nix, " ", suffix)
		
		player_mem.order_sel_n = nix
		build_menu_objects(player,true,true)
	end
end

script.on_event(defines.events.on_gui_click, on_gui_click)

--------------------------------------------------------------------------------------
local function on_gui_text_changed(event)
	local player = game.players[event.player_index]
	local event_name = event.element.name
	local prefix = string.sub(event_name,1,15)
	local nix = tonumber(string.sub(event_name,16,17))
	local suffix = string.sub(event_name,18)

	-- debug_print( "on_gui_text_changed ", player.name, " ", event_name )
	
	if prefix == "txt_blkmkt_chc_" then -- change order count
		local player_mem = global.player_mem[player.index]
		local chest = player_mem.opened_chest
		
		-- debug_print(nix, " ", suffix, " ", chest.orders[nix].count)
		
		local count = tonumber(event.element.text)
		if count ~= nil then
			chest.orders[nix].count = count
			compute_chest_data(chest)
			update_menu_chest(player,player_mem,false)
		end
	end
end

script.on_event(defines.events.on_gui_text_changed,on_gui_text_changed)

--------------------------------------------------------------------------------------
local interface = {}

function interface.reset()
	debug_print( "reset" )
	
	for _,force in pairs(game.forces) do
		force.reset_recipes()
		force.reset_technologies()
		local force_mem = global.force_mem[force.name]
		force_mem.credits = n
		force_mem.pause = false 
		force_mem.credits = 0
		force_mem.sales = 0
		force_mem.sales_taxes = 0
		force_mem.purchases = 0
		force_mem.purchases_taxes = 0
		force_mem.tax_rate = 0
		force_mem.transactions = {} 
		
		compute_force_data(force_mem)
		
		for _, chest in pairs(force_mem.chests_sell) do
			init_chest(chest)
			compute_chest_data(chest)
		end
		
		for _, chest in pairs(force_mem.chests_buy) do
			init_chest(chest)
			compute_chest_data(chest)
		end
	end

	for _, player in pairs(game.players) do
		if player.gui.top.flw_blkmkt then player.gui.top.flw_blkmkt.destroy() end
		init_player(player)
		local player_mem = global.player_mem[player.index]
		update_menu_gen(player,player_mem)
		update_menu_chest(player,player_mem,false)
	end
end

function interface.prices()
	debug_print( "prices" )
	
	update_groups()
	update_techs_costs()
	update_objects_prices_start()
end

function interface.credits(n)
	debug_print( "credits" )
	
	for _,force in pairs(game.forces) do
		local force_mem = global.force_mem[force.name]
		force_mem.credits = n
	end
	
	for _, player in pairs(game.players) do
		update_bar(player)
	end
end

remote.add_interface( "market", interface )

-- /c remote.call( "market", "reset" )
-- /c remote.call( "market", "prices" )
-- /c remote.call( "market", "credits", 1000000 )

